from passlib.context import CryptContext

pwd_context = CryptContext('bcrypt', deprecated='auto')


def verify_password(plain_password, hashed_password):
    return pwd_context.verify(plain_password, hashed_password)


def get_password_hash(password):
    return pwd_context.hash(password)



# TODO
# This file is incomplete need to be solution at scale while other connections on different host
# It's could be solve by using an database to store websocket connection under object type
# Corresponding to an uuid with only one connection
import logging
import uuid
from typing import Dict, Union

import requests
from fastapi import status, WebSocket, Path, Query
from starlette.websockets import WebSocketDisconnect

from configs.settings import app

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
formatter = logging.Formatter('%(asctime)s:%(levelname)s:%(module)s:%(lineno)d:%(message)s')
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)

# Create a connection list contains all all connection to specific endpoint
# Dictionary which have key as user_id, and value is websocket connection instance
GROUPS_ACTIVE_CONNECTION: Dict[Union[int, str], WebSocket] = {}
SYSTEM_ACTIVE_CONNECTION: Dict[Union[int, str], WebSocket] = {}


# TODO fix with rethink DB for horizontal scaling purpose

@app.websocket(path='/ws/group')
async def group_establish_connection(
        web_socket: WebSocket,
        user_id: Union[int, str] = Query(..., description='Use with access_token param for authentication purpose'),
        access_token: str = Query(..., description='Use for authentication purpose')
):
    global GROUPS_ACTIVE_CONNECTION
    # Checking if authentication success before accepting connection
    AUTHENTICATION_ENDPOINT = ''
    data = {
        'user_id': user_id,
        'access_token': access_token
    }
    response = requests.post(AUTHENTICATION_ENDPOINT, data=data)
    if response.status_code == status.HTTP_200_OK:
        conn_id = uuid.uuid4()
        await web_socket.send_json(
            {
                'conn_id': str(conn_id),
                'state': 'Connected'
            }
        )
        GROUPS_ACTIVE_CONNECTION[f'{conn_id}'] = web_socket
        try:
            json: Dict = await web_socket.receive_json()
            await web_socket.send_json(json)
        except WebSocketDisconnect:
            try:
                del GROUPS_ACTIVE_CONNECTION[f'{conn_id}']
                logger.info(f'Removed connection {conn_id}')
            except KeyError:
                logger.warning(f'{conn_id} does not exist to delete')
    await web_socket.close()


@app.websocket(path='/ws/system')
async def system_establish_connection(
        web_socket: WebSocket,
        user_id: Union[int, str] = Query(..., description='Use with access_token param for authentication purpose'),
        access_token: str = Query(..., description='Use for authentication purpose')
):
    global SYSTEM_ACTIVE_CONNECTION
    # Checking if authentication success before accepting connection
    AUTHENTICATION_ENDPOINT = ''
    data = {
        'user_id': user_id,
        'access_token': access_token
    }
    response = requests.post(AUTHENTICATION_ENDPOINT, data=data)
    if response.status_code == status.HTTP_200_OK:
        conn_id = uuid.uuid4()
        await web_socket.send_json(
            {
                'conn_id': str(conn_id),
                'state': 'Connected'
            }
        )
        SYSTEM_ACTIVE_CONNECTION[f'{conn_id}'] = web_socket
        try:
            json: Dict = await web_socket.receive_json()
            await web_socket.send_json(json)
        except WebSocketDisconnect:
            try:
                del SYSTEM_ACTIVE_CONNECTION[f'{conn_id}']
                logger.info(f'Removed connection {conn_id}')
            except KeyError:
                logger.warning(f'{conn_id} does not exist to delete')
    await web_socket.close()



import os
import subprocess
import shlex
import signal
import multiprocessing
from fastapi.middleware.cors import CORSMiddleware
from starlette.middleware.sessions import SessionMiddleware
from fastapi.openapi.utils import get_openapi
from pydantic import schema
from typing import Any
from pydantic.fields import ModelField
from configs.settings import (
    PROJECT_SECRET_KEY, PROJECT_TITLE, PROJECT_DESCRIPTION, __VERSION__, MONGO_CLIENT)
# Starting import api views
from app.api_views.hello import *


# Ending import api views
# API Versioning for FastAPI application
from fastapi_versioning import VersionedFastAPI
app = VersionedFastAPI(
    app, version_format='{major}.{minor}',
    prefix_format='/v{major}.{minor}',
    enable_latest=True
)


def start_celery_beat():
    # Starting celery schedule
    if not os.path.exists("./logs"):
        os.mkdir("./logs")
    if os.path.exists("./celerybeat-schedule"):
        os.remove("./celerybeat-schedule")
    celery_logs = open("logs/celery_tasks.log", "w", encoding="utf8")
    cmd_start_celery_schedule = 'celery -A app.tasks.celery worker -E -l info -B'
    started_celery_schedule = subprocess.Popen(
        shlex.split(cmd_start_celery_schedule),
        universal_newlines=True, stdout=celery_logs
    )
    started_celery_schedule.communicate()


@app.on_event('startup')
async def startup():
    start_celery_beat_pc = multiprocessing.Process(target=start_celery_beat)
    start_celery_beat_pc.start()
    # Save pid for later shutdown action
    MONGO_CLIENT['celery']['pid'].update_one({'start_celery_beat_pc': True}, {
                                             '$set': {'pid': start_celery_beat_pc.pid}}, upsert=True)


@app.on_event('shutdown')
async def shutdown():
    pid = MONGO_CLIENT['celery']['pid'].find_one(
        {'start_celery_beat_pc': True}, {'pid': 1}).get('pid')
    os.kill(pid, signal.SIGTERM)


def custom_openapi():
    openapi_schema = get_openapi(
        title=PROJECT_TITLE,
        description=PROJECT_DESCRIPTION,
        version=__VERSION__,
        routes=app.routes
    )
    openapi_schema['info']['x-logo'] = {
        'url': '{{PROJECT_LOGO_URL}}'
    }
    app.openapi_schema = openapi_schema
    return app.openapi_schema


app.openapi = custom_openapi

# Provide which origins should be accepted to call apis
origins = [
    'http://localhost',
    'http://127.0.0.1',
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.add_middleware(
    # Using session in request like
    # from fastapi.requests import Request
    # @app.get(...)
    # async def api_function(request: Resquest):
    #       request.sesssion['key'] = value
    SessionMiddleware, secret_key=PROJECT_SECRET_KEY
)

# Make custom field_schema to
# Exclude some field out of model when generating schema


def field_schema(field: ModelField, **kwargs: Any) -> Any:
    if field.field_info.extra.get("hidden_from_schema", False):
        raise schema.SkipField(f"{field.name} field is being hidden")
    else:
        return original_field_schema(field, **kwargs)


original_field_schema = schema.field_schema
schema.field_schema = field_schema



import subprocess
import shlex
import jwt
from datetime import datetime, timedelta
from typing import Optional
from fastapi import status, HTTPException, Header

from configs.settings import MONGO_CLIENT, DB_NAME


ALGORITHM = "HS256"
ACCESS_TOKEN_EXPIRE_MINUTES = 3000


def get_secret_key():
    cmd = "openssl rand -hex 16"
    completed_process = subprocess.run(shlex.split(cmd), capture_output=True)
    return completed_process.stdout.decode(encoding='utf-8').strip('\n')


def create_access_token(data: dict, expires_delta: Optional[timedelta] = None, secret_key=None):
    """Return JWT and it's secret key for decoding
    Args:
        data (dict): data to be encoded
        expires_delta (timedelta): Specify if there's limited time for token life
        secret_key (str): secret_key for encoding/decoding JWT token
    Returns:
        tuple: contains encoded_jwt, SECRET_KEY
    """
    SECRET_KEY = get_secret_key()
    to_encode = data.copy()
    if expires_delta:
        expire = datetime.utcnow() + expires_delta
    else:
        expire = datetime.utcnow() + timedelta(minutes=ACCESS_TOKEN_EXPIRE_MINUTES)
    to_encode.update({"exp": expire})
    if not secret_key:
        encoded_jwt = jwt.encode(to_encode, SECRET_KEY, algorithm=ALGORITHM)
        return encoded_jwt, SECRET_KEY
    encoded_jwt = jwt.encode(to_encode, secret_key, algorithm=ALGORITHM)
    return encoded_jwt, secret_key


def is_expired(encode_jwt, SECRET_KEY):
    """Check if token is expired or not
    Args:
        encode_jwt (str): JWT token for checking
        SECRET_KEY (str): Key for decoding JWT token
    Returns:
        bool: True if not expired, False otherwise
    """
    try:
        jwt.decode(encode_jwt, SECRET_KEY, algorithms=ALGORITHM)
    except Exception:
        return True
    return False


def get_data_from_access_token(encode_jwt, SECRET_KEY):
    """Check if token is expired or not
        Args:
            encode_jwt (str): JWT token for checking
            SECRET_KEY (str): Key for decoding JWT token
        Returns:
            dict: data from access token
    """
    return jwt.decode(encode_jwt, SECRET_KEY, algorithms=ALGORITHM)


async def valid_access_token(
        Authorization: str = Header(...,
                                    description='access_token return by login'),
        s_key: str = Header(..., description='secret_key return by login'),
):
    """Check if access token valid for accessing api
    Args:
        Bearer (str): Bearer token
        s_key (str): Secret key
    Returns:
        dict: Containing keys value which relevant to security session
    Examples:
        return {
            'encode_jwt': token,
            'type': bearer,
            's_key': s_key
        }
    Raises:
        HTTPException: if token is not valid
    """
    try:
        bearer, access_token = Authorization.split(' ')
        if not is_expired(encode_jwt=access_token, SECRET_KEY=s_key):
            data = get_data_from_access_token(
                encode_jwt=access_token, SECRET_KEY=s_key)
            email = data.get('email')
            user = MONGO_CLIENT[f'{DB_NAME}']['users'].find_one(
                {'email': {'$eq': email}})
            for _token in user.get('tokens'):
                if _token.get('access_token') == access_token:
                    token_type = _token.get('token_type')
                    if token_type != bearer:
                        raise HTTPException(
                            status_code=status.HTTP_400_BAD_REQUEST, detail='Token type is not valid')
                    return {
                        'encode_jwt': access_token,
                        'type': bearer,
                        's_key': s_key
                    }
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST, detail='Token is not valid')
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail='Token type is expired')



import git
import os
import subprocess
import shutil
import shlex
import re
from pathlib import Path
from fileinput import FileInput
from distutils.dir_util import copy_tree

CWD = Path(__file__).parent

LOCAL_DIR = Path(__file__).parent / "temp"

GIT_REPO = LOCAL_DIR / ".git"


def remove_local_dir():
    if os.path.isdir(LOCAL_DIR.as_posix()):
        shutil.rmtree(LOCAL_DIR.as_posix())


remove_local_dir()

REMOTE_REPO = "https://github.com/kampvie/fastapi_project_pattern.git"

BRANCH = "main"


def start_creation(path: Path = None, use_local_posgres: bool = False, use_remote_postgres_url: bool = False,
    use_circle_ci: bool = False, use_gitlab_ci: bool = False
):
    """Create a project template
    Args:
        path (Path): Location a project should be placed once created
        use_local_posgres (bool): Whether use local PostgreSQL or not
        use_remote_postgres_url (bool): Use remote PostgreSQL or not
        use_circle_ci (bool): Using Circle CI/CD
        use_gitlab_ci (bool): Using Gitlab CI/CD
    """
    if not path:
        raise ValueError("path parameter is required")
    # Clone git repo into local
    git.Repo.clone_from(REMOTE_REPO, LOCAL_DIR.as_posix(), branch=BRANCH)

    cmd = shlex.split("openssl rand -hex 32")

    SECRET_KEY = subprocess.run(
        cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE).stdout.decode("utf-8").strip()

    VARS = {
        # KEY: [DefaultValue, Promtline, Required, ValidateRegrexPattern, Type]
        'PROJECT_NAME': ["", "Project name >>> ", True, None, str],

        'PROJECT_LOGO_URL': ["", "Project logo url >>> ", True, None, str],

        'PROJECT_GIT_SSH_REPO': ["", "SSH git repository >>> ", True, None, str],

        'PROJECT_TITLE': ["", "Title >>> ", True, None, str],

        'PROJECT_SECRET_KEY': [SECRET_KEY, f"Secret[Default: {SECRET_KEY}] >>> ", False, None, str],

        'PROJECT_DESCRIPTION': [None, "Description >>> ", True, None, str],

        'PORT': ["2306", "Port to run [0-65536](Default: 2306) >>> ", False, None, str],

        'CLOUDINARY_URL': ["None", "CLOUDINARY_URL(Optional) >>> ", False, None, str],

        'SENDGRID_API_KEY': ["None", "SENDGRID_API_KEY(Optional) >>> ", False, None, str],

        'SENDGRID_FROM_EMAIL': ["None",
            "SENDGRID_FROM_EMAIL(Optional), email to send others  >>> ",
            False,
            None,
            str
        ],

        'LOCAL_POSTGRES_USER': ["None", "POSTGRES_USER[Optional] >>> ", False, None, str],

        'LOCAL_POSTGRES_DB': ["None", "POSTGRES_DB[Optional] >>> ", False, None, str],

        'LOCAL_POSTGRES_PASSWORD': ["None", "POSTGRES_PASSWORD[Optional] >>> ", False, None, str],

        'REMOTE_POSTGRES_URL': ["None", "REMOTE_POSTGRES_URL[Optional] >>> ", False, None, str],

        'MONGO_INITDB_ROOT_USERNAME': ["mongo", "MONGO_INITDB_ROOT_USERNAME(Default: mongo) >>> ", False, None, str],

        'MONGO_INITDB_ROOT_PASSWORD': ["mongo", "MONGO_INITDB_ROOT_PASSWORD >>> ", True, None, str],

        'MONGO_DATABASE_NAME': ["master", "MONGO_DATABASE_NAME(Default: master) >>>", True, None, str],

        'REMOTE_MONGO_URL': ["None", "REMOTE_MONGO_URL >>> ", True, None, str],

        'RABBITMQ_DEFAULT_USER': ["rabbit", "RABBITMQ_DEFAULT_USER(Default: rabbit) >>> ", False, None, str],

        'RABBITMQ_DEFAULT_PASS': ["rabbit", "RABBITMQ_DEFAULT_PASS >>> ", True, None, str],
    }

    # Check use_local_posgres flag. If it was not used then remove out of VARS 
    if not use_local_posgres:
        del VARS['LOCAL_POSTGRES_USER']
        del VARS['LOCAL_POSTGRES_DB']
        del VARS['LOCAL_POSTGRES_PASSWORD']
    if not use_remote_postgres_url:
        del VARS['REMOTE_POSTGRES_URL']

    def is_valid(v, text):
        # Check if text is empty but this field is required
        # So with that said it's not valid
        if len(text) == 0 and v[2]:
            return False
        # TODO Check regrex to validate pattern
        return True

    for k, v in VARS.items():
        while True:
            text = input(v[1]).strip()
            # Project name is not allowed contains spaces
            if k == "PROJECT_NAME" and re.search(r"\s", text):
                continue
            if is_valid(v, text):
                v[0] = text or v[0]
                break

    def replace(file_path: Path, text: str, subs: str, flags=0):
        """Replace content in a file
        """
        # Opening file using FileInput
        with FileInput(file_path.as_posix(), inplace=True) as f:
            for line in f:
                print(line.replace(text, subs), end='')

    # Remove .git in temp repo
    shutil.rmtree(GIT_REPO.as_posix())

    def change_files(path: Path):
        for file in os.listdir(path.as_posix()):
            _path = path / file
            if os.path.isfile(_path.as_posix()):
                for k, v in VARS.items():
                    replace(_path, "{{" + k + "}}", VARS[k][0])
                # Check to uncomment codeblock
                if use_local_posgres:
                    replace(_path, "#_p ", "")
            elif os.path.isdir(_path):
                change_files(_path)

    # Call function to apply
    change_files(LOCAL_DIR)

    # Remove redunrant files, folders
    redundancies = [
        LOCAL_DIR / "pyproject.toml",
        LOCAL_DIR / "LICENSE",
        LOCAL_DIR / "README.md",
    ]

    if not use_gitlab_ci:
        redundancies.append(
            LOCAL_DIR / ".gitlab-ci.yml"
        )

    if not use_circle_ci:
        CIRCLE_CI = LOCAL_DIR / ".circleci"
        shutil.rmtree(CIRCLE_CI.as_posix())

    for _path in redundancies:
        os.remove(_path.as_posix())

    FASTAPI_PROJECT_PATTERN = LOCAL_DIR / "fastapi_project_pattern"
    shutil.rmtree(FASTAPI_PROJECT_PATTERN.as_posix())

    # Create project directory and put code in it
    PROJECT_PATH: Path = CWD / f'{VARS["PROJECT_NAME"][0]}'
    path = path / f'{VARS["PROJECT_NAME"][0]}'

    os.mkdir(PROJECT_PATH.as_posix())
    copy_tree(LOCAL_DIR.as_posix(), PROJECT_PATH.as_posix())

    os.mkdir(path.as_posix())
    copy_tree(PROJECT_PATH.as_posix(), path.as_posix())

    shutil.rmtree(PROJECT_PATH.as_posix())
    remove_local_dir()



from typing import List
import datetime

from pydantic import BaseModel, EmailStr


class Token(BaseModel):
    access_token: str
    token_type: str
    device_id: str


class User(BaseModel):
    username: str
    tokens: List[Token]
    secret_key: str
    email: EmailStr
    hashed_password: str
    datetime_created: datetime.datetime
    is_verified: bool


from configs.settings import (DB_HOST, DB_PORT, MONGO_INITDB_ROOT_USERNAME, MONGO_INITDB_ROOT_PASSWORD)

result_backend = f'mongodb://{MONGO_INITDB_ROOT_USERNAME}:{MONGO_INITDB_ROOT_PASSWORD}@{DB_HOST}:{DB_PORT}'
result_persistent = True
mongodb_backend_settings = {
    'database': 'celery',
    'taskmeta_collection': f'celery_taskmeta',
}
task_reject_on_worker_lost = True


import os
from fastapi import FastAPI
from pymongo import MongoClient
from sqlalchemy import create_engine

PROJECT_SECRET_KEY = "{{PROJECT_SECRET_KEY}}"

DEBUG = os.environ.get('DEBUG', 0)

PROJECT_TITLE = "{{PROJECT_TITLE}}"
PROJECT_DESCRIPTION = "{{PROJECT_DESCRIPTION}}"
__VERSION__ = '1.0'

# Initial FastAPI app
app = FastAPI(
    debug=bool(DEBUG),
    title=PROJECT_TITLE,
    docs_url=None,
    redoc_url=None
)

# Get DB host from docker-compose environment
DB_HOST = os.environ.get('DB_HOST')

DB_PORT = os.environ.get('DB_PORT')

RABBITMQ = os.environ.get('RABBITMQ')

CLOUDINARY_URL = os.environ.get('CLOUDINARY_URL')

SENDGRID_API_KEY = os.environ.get('SENDGRID_API_KEY')

SENDGRID_FROM_EMAIL = os.environ.get('SENDGRID_FROM_EMAIL')

REMOTE_MONGO_URL = os.environ.get('REMOTE_MONGO_URL')

MONGO_INITDB_ROOT_USERNAME = os.environ.get('MONGO_INITDB_ROOT_USERNAME')

MONGO_INITDB_ROOT_PASSWORD = os.environ.get('MONGO_INITDB_ROOT_PASSWORD')

REMOTE_POSTGRES_URL = os.environ.get('REMOTE_POSTGRES_URL')

RABBITMQ_DEFAULT_USER = os.environ.get('RABBITMQ_DEFAULT_USER')

RABBITMQ_DEFAULT_PASS = os.environ.get('RABBITMQ_DEFAULT_PASS')

CELERY_APP_NAME = 'app'

MONGO_CLIENT = None

# Database name to use it as MONGO_CLIENT[f'{DB_NAME}']
DB_NAME = os.environ.get('MONGO_DATABASE_NAME')

if all([DB_HOST, DB_PORT]):
    # initial connection to database
    MONGO_CLIENT = MongoClient(
        f'mongodb://{MONGO_INITDB_ROOT_USERNAME}:{MONGO_INITDB_ROOT_PASSWORD}@{DB_HOST}:{DB_PORT}/?authMechanism=SCRAM-SHA-256')
    # Create schedule db for storing result backend
    if not MONGO_CLIENT['celery'].command('usersInfo', usersInfo={"user": MONGO_INITDB_ROOT_USERNAME, "db": "celery"}).get('users'):
        MONGO_CLIENT['celery'].command(
            'createUser',
            createUser=MONGO_INITDB_ROOT_USERNAME,
            pwd=MONGO_INITDB_ROOT_PASSWORD,
            roles=["readWrite", "dbAdmin"],
            mechanisms=["SCRAM-SHA-256"],
        )

if REMOTE_MONGO_URL and REMOTE_MONGO_URL != "None":
    # Switch back to remote mongo database service
    MONGO_CLIENT = MongoClient(REMOTE_MONGO_URL)

# Establishing Connectivity with SQL Database
POSTGRES_HOST = os.environ.get('POSTGRES_HOST')
POSTGRES_PORT = os.environ.get('POSTGRES_PORT')
POSTGRES_USER = os.environ.get('POSTGRES_USER')
POSTGRES_DB = os.environ.get('POSTGRES_DB')
POSTGRES_PASSWORD = os.environ.get('POSTGRES_PASSWORD')

SQL_ENGINE = None

if all([POSTGRES_USER, POSTGRES_DB, POSTGRES_PASSWORD]):
    SQL_ENGINE = create_engine(
        f"postgresql://{POSTGRES_USER}:{POSTGRES_PASSWORD}@{POSTGRES_HOST}:{POSTGRES_PORT}/{POSTGRES_DB}"
    )



# Define all tags for api
TAGS_METADATA = [
    {
        'name': 'greet',
        'description': 'Greeting API',
        'externalDocs': {
            'description': 'None',
            'url': 'https://external-docs'
        }
    },
    {
        'name': 'account',
        'description': 'Các API liên quan tài khoản người dùng',
        'externalDocs': {
            'description': 'None',
            'url': 'https://external-docs'
        }
    }
]



import logging


def get_stream_logger(name):
    """Get logger and stream output to stdin
    """
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter(
        '%(asctime)s:%(levelname)s:%(module)s:%(lineno)d:%(message)s'
    )
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    return logger



from typing import List, Dict, Optional

from pydantic import BaseModel, Field

from models.user import Token


class CreateAccountResponse200(BaseModel):
    status: str = "Tạo tài khoản thành công!"
    access_token: str = Field(
        ...,
        description='JWT Access token'
    )
    device_id: str = Field(..., description='Id thiết bị')
    secret_key: str = Field(..., description='Secret key cho tài khoản')
    is_verified: bool = Field(
        False, description='Trạng thái xác thực tài khoản')


class CreateAccountResponse403(BaseModel):
    status: str = "Lỗi"
    msg: str = "Email đã tồn tại"


class LoginResponse200(BaseModel):
    token: Token = Field(..., description='Access token')
    secret_key: str = Field(..., description='Secret key')


class LoginResponse403(BaseModel):
    status: str = Field(default="Lỗi")
    msg: Optional[str] = Field(description='Thông báo lỗi')


class LoginResponse404(BaseModel):
    status: str = Field(default="Lỗi")
    msg: Optional[str] = Field(description='Không tồn tại')


class UpdateAccount200(BaseModel):
    status: str = Field(default='Cập nhật tài khoản thành công!')


class UpdateAccount302(BaseModel):
    status: str = Field(default='Lỗi!')
    msg: str


class UpdateAccount400(BaseModel):
    status: str = Field(default='Lỗi!')


class VerifyAccountResponse200(BaseModel):
    status: str = Field(default='Thành công!')
    msg: str = Field(default='Đã xác nhận tài khoản!')


class VerifyAccountResponse403(BaseModel):
    status: str = Field(default='Lỗi!')
    msg: str = Field(default='Key xác nhận không hợp lệ hoặc hết thời gian')


class ResetPasswordResponse201(BaseModel):
    status: str = Field(default='Thành công!')
    msg: str = Field(default='Vui lòng kiểm tra email và nhập mã!')


class ResetPasswordResponse404(BaseModel):
    status: str = Field(default='Lỗi!')
    msg: str = Field(default='Tài khoản không tồn tại!')


class PutResetPasswordResponse200(BaseModel):
    status: str = Field(default='Thành công!')
    msg: str = Field(default='Mật khẩu đã được thiết lập lại')


class PutResetPasswordResponse400(BaseModel):
    status: str = Field(default='Lỗi!')
    msg: str = Field(default='Mã hết thời gian hoặc không tồn tại!')


class ChangePasswordResponse200(BaseModel):
    status: str = Field(default='Thành công!')
    msg: str = Field(default='Gửi yêu cầu thay đổi mật khẩu thành công!')
    nonce: int = Field(default='123456', description='Mã 6 chữ số dùng 1 lần')


class UpdatePasswordResponse200(BaseModel):
    status: str = Field(default='Thành công!')
    msg: str = Field(default='Cập nhật mật khẩu thành công!')


class UpdatePasswordResponse400(BaseModel):
    status: str = Field(default='Lỗi!')
    msg: str = Field(default='Mật khẩu không khớp!')


class UpdatePasswordResponse403(BaseModel):
    status: str = Field(default='Lỗi!')
    msg: str = Field(..., description='Thông tin lỗi')



# Start application before running pytest
from time import sleep

import pytest
import requests

from tests.settings import proc


@pytest.fixture(scope='function')
def supply_text_params():
    text = "Welcome to our test for branch new project"
    return text

def test_hello(supply_text_params):
    proc.start()
    sleep(1)
    response = requests.get(f'http://0.0.0.0:2306/hello/?greetings={supply_text_params}')
    assert response.status_code == 200
    proc.kill()



from configs.celery import app_celery
from app.utils.logger import get_stream_logger

logger = get_stream_logger(__name__)

@app_celery.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):
    sender.add_periodic_task(
        30,
        example.s(),
        name='Hello every 30 seconds'
    )

@app_celery.task
def example():
    logger.info("Example periodic task -- Hello")


from typing import Optional

from fastapi import status, Query
from fastapi.responses import JSONResponse
from fastapi_versioning import version
from configs.settings import app


@app.get(path='/hello', tags=['greet'])
@version(1, 0)
async def hello(greeting: Optional[str] = Query(default='Hello world')):
    return JSONResponse(content={'status': greeting + ' Version 1.0'}, status_code=status.HTTP_200_OK)


@app.get(path='/hello', tags=['greet'])
@version(1, 1)
async def hello(greeting: Optional[str] = Query(default='Hello world')):
    return JSONResponse(content={'status': greeting + ' Version 1.1'}, status_code=status.HTTP_200_OK)



# using SendGrid's Python Library
# https://github.com/sendgrid/sendgrid-python
import logging
import os
from sendgrid import SendGridAPIClient
from sendgrid.helpers.mail import Mail


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
formatter = logging.Formatter(
    '%(asctime)s:%(levelname)s:%(module)s:%(lineno)d:%(message)s'
)
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)


async def create_account_verification(from_email, to_emails, data: dict):
    """Gửi thông báo xác nhận tài khoản mới tạo
    Args:
        from_email (str): Email gửi
        to_email (list, tuple, str): Email nhận hoặc danh sách email nhận

    Returns:
        True if sent successfully else False
    """
    message = Mail(
        from_email=from_email,
        to_emails=to_emails)
    message.dynamic_template_data = data
    message.template_id = "d-bfe931e480ef4069acf79b2bb5aa76fe" #TODO thay thế template id bằng dynamic template id trên sendgrid
    try:
        sg = SendGridAPIClient(os.environ.get('SENDGRID_API_KEY'))
        response = sg.send(message)
        logger.info(response.status_code)
        logger.info(response.body)
        logger.info(response.headers)
    except Exception as e:
        logger.info(e.message)
        return False
    return True


async def send_reset_password_email(from_email, to_emails, data: dict):
    """Gửi email reset mật khẩu tài khoản
    Args:
        from_email (str): Email gửi
        to_email (list, tuple, str): Email nhận hoặc danh sách email nhận

    Returns:
        True if sent successfully else False
    """
    message = Mail(
        from_email=from_email,
        to_emails=to_emails)
    message.dynamic_template_data = data
    message.template_id = "d-e1601fc60c5c4d74b7af28e24006950b" #TODO thay thế template id bằng dynamic template id trên sendgrid
    try:
        sg = SendGridAPIClient(os.environ.get('SENDGRID_API_KEY'))
        response = sg.send(message)
        logger.info(response.status_code)
        logger.info(response.body)
        logger.info(response.headers)
    except Exception as e:
        logger.info(e.message)
        return False
    return True



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['fastapi_project_pattern']

package_data = \
{'': ['*']}

install_requires = \
['GitPython >= 3.1.24']

setup(name='fastapi_project_pattern',
      version='1.1.8',
      description='This project was created as a shortcut for fastapi project creation',
      author=None,
      author_email='Le Tanh Sang <sang.tanhle@gmail.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
     )



import os

from celery import Celery
from configs.settings import (
    CELERY_APP_NAME, RABBITMQ_DEFAULT_USER, RABBITMQ_DEFAULT_PASS)

RABBITMQ = os.environ.get('RABBITMQ')
# Initialize an app for celery task queue
# Where app is up-level package and tasks is the sub-package for store and calling task
app_celery = Celery(f'{CELERY_APP_NAME}',
                    broker=f'amqp://{RABBITMQ_DEFAULT_USER}:{RABBITMQ_DEFAULT_PASS}@{RABBITMQ}:5672',
                    include=['app.tasks.celery'])
# Optional configuration
app_celery.config_from_object('configs.celeryconfig')



import logging
import random
from app.utils.send_verification_email import create_account_verification, send_reset_password_email
from fastapi import status, Form, Body, Depends, Path, Query
from fastapi.encoders import jsonable_encoder
from bson.objectid import ObjectId
from fastapi.responses import JSONResponse
from pydantic import EmailStr
from pymongo import ReturnDocument
from configs.settings import DB_NAME, PROJECT_SECRET_KEY, app, MONGO_CLIENT, SENDGRID_FROM_EMAIL
from secure._token import get_data_from_access_token, create_access_token, valid_access_token, is_expired
from secure._password import get_password_hash, verify_password

from models.response.account import (
    CreateAccountResponse200,
    CreateAccountResponse403,
    LoginResponse200,
    LoginResponse403,
    LoginResponse404,
    VerifyAccountResponse200,
    VerifyAccountResponse403,
    ResetPasswordResponse201,
    ResetPasswordResponse404,
    PutResetPasswordResponse200,
    PutResetPasswordResponse400,
    ChangePasswordResponse200,
    UpdatePasswordResponse200,
    UpdatePasswordResponse400,
    UpdatePasswordResponse403,
)

from models.user import *

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
formatter = logging.Formatter(
    '%(asctime)s:%(levelname)s:%(module)s:%(lineno)d:%(message)s'
)
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)


@app.post(path='/account', responses={
    status.HTTP_200_OK: {
        'model': CreateAccountResponse200
    },
    status.HTTP_403_FORBIDDEN: {
        'model': CreateAccountResponse403
    }
}, tags=['account'])
async def create_account(
    email: EmailStr = Form(..., description='Email'),
    username: str = Form(..., description='Username'),
    password: str = Form(..., description='Password'),
    device_id: str = Form(..., description='Device ID'),
    verification_link: str = Form(...,
                                  description='Link giao diện xác nhận tài khoản')
):
    """Tạo tài khoản hệ thống"""
    import datetime
    # Check if user already exist
    if MONGO_CLIENT[f'{DB_NAME}']['users'].find({"email": {"$eq": email}}).count():
        return JSONResponse(content={
            "status": "Lỗi",
            "msg": "Email đã tồn tại!"
        }, status_code=status.HTTP_403_FORBIDDEN)
    # Create user in db
    access_token, secret_key = create_access_token(
        data={
            'email': email,
            'username': username
        }
    )
    token = Token(
        access_token=access_token,
        token_type='Bearer',
        device_id=device_id
    )
    user = User(
        username=username,
        tokens=[token],
        secret_key=secret_key,
        email=email,
        hashed_password=get_password_hash(password),
        datetime_created=datetime.datetime.now(),
        is_verified=False
    )
    inserted_id = MONGO_CLIENT[f'{DB_NAME}']['users'].insert_one(
        jsonable_encoder(user)
    ).inserted_id
    # Send email to verify account
    # Create access token container userId and expire time for a link
    encoded_jwt, _ = create_access_token({'user_id': str(inserted_id)},
                                         expires_delta=datetime.timedelta(minutes=30), secret_key=PROJECT_SECRET_KEY)
    await create_account_verification(SENDGRID_FROM_EMAIL, email, data={
        'User_Name': username,
        'msg': 'Nhấn vào link hoặc nút xác nhận bên dưới để xác nhận tài khoản',
        'Verification_Link': f'{verification_link}?verifyKey={encoded_jwt}'
    })
    return JSONResponse(content={
        'status': 'Tạo tài khoản thành công',
        'access_token': access_token,
        'secret_key': secret_key,
        'device_id': device_id,
        'is_verified': False
    }, status_code=status.HTTP_200_OK)


@app.post(path='/account/login', responses={
    status.HTTP_200_OK: {
        'model': LoginResponse200
    },
    status.HTTP_403_FORBIDDEN: {
        'model': LoginResponse403
    },
    status.HTTP_404_NOT_FOUND: {
        'model': LoginResponse404
    }
}, tags=['account'])
async def login(email: EmailStr = Body(...), password: str = Body(...), device_id: str = Body(..., description='Id của thiết bị đăng nhập')):
    """Đăng nhập"""
    user = MONGO_CLIENT[f'{DB_NAME}']['users'].find_one(
        {'email': {'$eq': email}}
    )
    if not user:
        return JSONResponse(content={'status': 'Lỗi', 'msg': 'Không tồn tại!'}, status_code=status.HTTP_404_NOT_FOUND)
    if verify_password(password, user.get('hashed_password')):
        # Create new access token for device
        secret_key = user.get('secret_key')
        access_token, _ = create_access_token(
            data={
                'email': user.get('email'),
                'username': user.get('username')
            },
            secret_key=secret_key
        )
        token = Token(
            access_token=access_token,
            token_type='Bearer',
            device_id=device_id
        )
        # If device id exist we just need to update toke for only that device
        is_existed = False
        for _token in user.get('tokens'):
            if _token.get('device_id') == device_id:
                is_existed = True

        if is_existed:
            MONGO_CLIENT[f'{DB_NAME}']['users'].find_one_and_update(
                filter={
                    '$and': [
                        {'email': {'$eq': email}},
                        {'tokens.device_id': {'$eq': device_id}}
                    ]
                },
                update={
                    '$set': {'tokens.$': jsonable_encoder(token)}
                }
            )
            return JSONResponse(content={'token': jsonable_encoder(token), 'secret_key': secret_key},
                                status_code=status.HTTP_200_OK)
        else:
            if len(user.get('tokens')) >= 5:
                return JSONResponse(content={'status': 'Lỗi', 'msg': 'Đạt giới hạn 5 thiết bị đăng nhập'}, status_code=status.HTTP_403_FORBIDDEN)
            user = MONGO_CLIENT[f'{DB_NAME}']['users'].find_one_and_update(
                filter={'email': {'$eq': email}},
                update={
                    '$push': {
                        'tokens': jsonable_encoder(token)
                    }
                },
                return_document=ReturnDocument.AFTER
            )
            return JSONResponse(content={'token': jsonable_encoder(token), 'secret_key': secret_key},
                                status_code=status.HTTP_200_OK)
    return JSONResponse(content={'status': 'Lỗi', 'msg': 'Đăng nhập không thành công!'}, status_code=status.HTTP_403_FORBIDDEN)


@app.get(path='/account/verify', responses={
    status.HTTP_200_OK: {
        'model': VerifyAccountResponse200
    },
    status.HTTP_403_FORBIDDEN: {
        'model': VerifyAccountResponse403
    }
}, tags=['account'])
async def verify_account(verify_key: str = Query(..., description='Key xác nhận tài khoản được gửi kèm là verifyKey trong link xác nhận tài khoản')):
    if not is_expired(verify_key, PROJECT_SECRET_KEY):
        data = get_data_from_access_token(verify_key, PROJECT_SECRET_KEY)
        user_id = data.get('user_id')
        _id = ObjectId(user_id)
        MONGO_CLIENT[f'{DB_NAME}']['users'].find_one_and_update(
            filter={'_id': _id},
            update={'$set': {'is_verified': True}}
        )
        return JSONResponse(content={'status': 'Thành công!', 'msg': 'Đã xác nhận tài khoản!'}, status_code=status.HTTP_200_OK)
    else:
        return JSONResponse(content={'status': 'Lỗi!', 'msg': 'Key xác nhận không hợp lệ hoặc hết thời gian'}, status_code=status.HTTP_403_FORBIDDEN)


@app.post(path='/account/password/reset', responses={
    status.HTTP_201_CREATED: {
        'model': ResetPasswordResponse201
    },
    status.HTTP_404_NOT_FOUND: {
        'model': ResetPasswordResponse404
    }
}, tags=['account'])
async def reset_password(
    email: EmailStr = Form(..., description='Email khôi phục mật khẩu'),
):
    """ 
        Khôi phục lại mật khẩu, hệ thống gửi mail kèm theo mã để khôi phục lại mật khẩu
    """
    import datetime
    import secrets
    keyonce = secrets.token_urlsafe(6)
    expire_in = datetime.datetime.now() + datetime.timedelta(minutes=30)
    expire_at_timestamp = expire_in.timestamp()
    user = MONGO_CLIENT[f'{DB_NAME}']['users'].find_one(
        filter={'email': {'$eq': email}}
    )
    if not user:
        return JSONResponse(content={'status': 'Lỗi!', 'msg': 'Tài khoản không tồn tại!'}, status_code=status.HTTP_404_NOT_FOUND)
    # Check if keyonce is not used
    await send_reset_password_email(SENDGRID_FROM_EMAIL, email, data={
        'msg': 'Vui lòng xác nhận địa chỉ email của bạn bằng cách nhập mã xác minh bên dưới. Mã có giá trị trong 30 phút',
        'keyonce': keyonce
    })
    MONGO_CLIENT[f'{DB_NAME}']['users'].find_one_and_update(
        filter={'email': {'$eq': email}},
        update={'$set': {'keyonce': keyonce,
                         'keyonce_expire_at': expire_at_timestamp}}
    )
    return JSONResponse(content={'status': 'Thành công!', 'msg': 'Vui lòng kiểm tra email và nhập mã!'}, status_code=status.HTTP_201_CREATED)


@app.put(path='/account/password/reset', responses={
    status.HTTP_200_OK: {
        'model': PutResetPasswordResponse200
    },
    status.HTTP_400_BAD_REQUEST: {
        'model': PutResetPasswordResponse400
    }
}, tags=['account'])
async def reset_password(
    keyonce: str = Form(..., description='Mã một lần để khôi phục mật khẩu'),
    password: str = Form(..., description='Mật khẩu lần một'),
    re_password: str = Form(..., description='Mật khẩu lần hai'),
):
    """Khôi phục lại mật khẩu"""
    if password != re_password:
        return JSONResponse(content={'status': 'Lỗi!', 'msg': 'Mật khẩu không khớp!'}, status_code=status.HTTP_400_BAD_REQUEST)
    user = MONGO_CLIENT[f'{DB_NAME}']['users'].find_one_and_update(
        filter={
            'keyonce': {'$eq': keyonce}, 'keyonce_expire_at': {'$gte': datetime.datetime.now().timestamp()}
        },
        update={
            '$set': {
                'tokens': [],
                'hashed_password': get_password_hash(password)
            },
            '$unset': {
                'keyonce': '',
                'keyonce_expire_at': ''
            }
        }
    )
    if not user:
        return JSONResponse(content={'status': 'Lỗi!', 'msg': 'Mã hết thời gian hoặc không tồn tại!'}, status_code=status.HTTP_400_BAD_REQUEST)
    return JSONResponse(content={'status': 'Thành công!', 'msg': 'Mật khẩu đã được thiết lập lại!'}, status_code=status.HTTP_200_OK)


@app.get(path='/account/password/change', responses={
    status.HTTP_200_OK: {
        'model': ChangePasswordResponse200
    }
}, tags=['account'])
async def change_password(
    valid_access_token: dict = Depends(valid_access_token)
):
    """Gửi yêu cầu thay đổi mật khẩu"""
    import datetime
    s_key = valid_access_token.get('s_key')
    nonce = random.randint(100000, 999999)
    expire_in = datetime.datetime.now() + datetime.timedelta(minutes=30)
    expire_at_timestamp = expire_in.timestamp()
    user = MONGO_CLIENT[f'{DB_NAME}']['users'].find_one(
        filter={'secret_key': s_key})
    while user:
        if user.get('nonce_expire_at') and user.get('nonce_expire_at') < datetime.datetime.now().timestamp():
            MONGO_CLIENT[f'{DB_NAME}']['users'].update_one(filter={'secret_key': {'$eq': s_key}}, update={
                                                           '$unset': {'nonce': '', 'nonce_expire_at': ''}})
        nonce = random.randint(100000, 999999)
        user = MONGO_CLIENT[f'{DB_NAME}']['users'].find_one(
            filter={'nonce': nonce})
    MONGO_CLIENT[f'{DB_NAME}']['users'].find_one_and_update(
        filter={'secret_key': {'$eq': s_key}},
        update={'$set': {'nonce': nonce, 'nonce_expire_at': expire_at_timestamp}}
    )
    return JSONResponse(content={'status': 'Thành công!', 'msg': 'Gửi yêu cầu thay đổi mật khẩu thành công!', 'nonce': nonce}, status_code=status.HTTP_200_OK)


@app.post(path='/account/password/change', responses={
    status.HTTP_200_OK: {
        'model': UpdatePasswordResponse200
    },
    status.HTTP_400_BAD_REQUEST: {
        'model': UpdatePasswordResponse400
    },
    status.HTTP_403_FORBIDDEN: {
        'model': UpdatePasswordResponse403
    }
}, tags=['account'])
async def update_password(
    nonce: int = Form(..., description='Mã một lần gồm 6 chữ số'),
    o_password: str = Form(..., description='Mật khẩu khẩu cũ'),
    password: str = Form(..., description='Mật khẩu lần một'),
    re_password: str = Form(..., description='Mật khẩu lần hai'),
    valid_access_token: dict = Depends(valid_access_token)
):
    """Cập nhật mật khẩu"""
    if password != re_password:
        return JSONResponse(content={'status': 'Lỗi!', 'msg': 'Mật khẩu không khớp!'}, status_code=status.HTTP_400_BAD_REQUEST)
    user = MONGO_CLIENT[f'{DB_NAME}']['users'].find_one(
        {
            'nonce': {'$eq': nonce},
            'nonce_expire_at': {
                '$gte': datetime.datetime.now().timestamp()
            }
        },
        {'hashed_password': True}
    )
    if not user:
        return JSONResponse(content={'status': 'Lỗi', 'msg': 'Mã thay đổi mật khẩu không đúng hoặc hết hạn!'}, status_code=status.HTTP_403_FORBIDDEN)
    if not verify_password(o_password, user.get('hashed_password')):
        return JSONResponse(content={'status': 'Lỗi!', 'msg': 'Mật khẩu cũ chưa chính xác!'}, status_code=status.HTTP_403_FORBIDDEN)
    MONGO_CLIENT[f'{DB_NAME}']['users'].find_one_and_update(
        filter={'nonce': {'$eq': nonce}},
        update={
            '$set': {'hashed_password': get_password_hash(password)},
            '$unset': {'nonce': '', 'nonce_expire_at': ''}
        }
    )
    return JSONResponse(content={'status': 'Thành công!', 'msg': 'Cập nhật mật khẩu thành công!'}, status_code=status.HTTP_200_OK)
